#include "AWList.h"

#include <stdio.h>
#include <stdlib.h>

typedef enum { TRUE = 1, FALSE = 0 } BOOL;

// ================> SOURCE CODE BELOW <================
AWList* AWList_init(AWList* list) {
  if (list == NULL) {
    list = calloc(1, sizeof(AWList));

    list->free_me = TRUE;
  } else {
    list->free_me = FALSE;
  }

  list->head = calloc(1, sizeof(AWNode));
  list->tail = list->head;
  list->iter = list->head;
  list->cnt  = 0;
  return list;
}

void AWList_destroy(AWList* list) {
  AWList_clear(list);

  free(list->head);
  if (list->free_me) { free(list); }
}

void AWList_clear(AWList* list) {
  if (!list) { return; }
  for (; list->tail != list->head; AWList_delete_node(list, list->tail))
    ;
}

AWNode* AWNode_init(void* data, void (*delete_func)(void*)) {
  AWNode* node = malloc(sizeof(AWNode));
  node->data            = data;
  node->delete_func     = delete_func;

  node->prev = NULL;
  node->next = NULL;

  return node;
}

void AWNode_destroy(AWNode* node) {
  if (node->next) { node->next->prev = node->prev; }
  node->prev->next = node->next;

  if (node && node->delete_func) { node->delete_func(node->data); }
  free(node);
}

void AWList_delete_node(AWList* list, AWNode* node) {
  if (!list || !list->head || !node) return;

  if (list->tail == node) { list->tail = node->prev; }
  if (list->iter == node) { list->iter = node->prev; }

  AWNode_destroy(node);
  list->cnt -= 1;
}

void AWList_add_head(AWList* list, void* data, void (*delete_func)(void*)) {
  if (!list || !list->head) return;

  AWNode* node  = AWNode_init(data, delete_func);
  AWNode* dummy = list->head;

  node->prev = dummy;
  node->next = dummy->next;

  dummy->next->prev = node;
  dummy->next       = node;

  list->iter = node;
  list->cnt += 1;
}

void AWList_add_tail(AWList* list, void* data, void (*delete_func)(void*)) {
  AWNode* node = AWNode_init(data, delete_func);

  node->prev = list->tail;
  node->next = NULL;

  list->tail->next = node;
  list->tail       = node;

  list->cnt += 1;
}

AWNode* AWList_find_node(AWList* list,
                         void*   data,
                         int (*cmp_callback)(void*, void*)) {
  AWNode* node;
  AWList_for_each(node, list) {
    if (cmp_callback(node->data, data) == 0) return node;
  }

  return NULL;
}

AWNode *AWList_iterator(AWList *list)
{
	if (!list || list->iter == list->tail->next)
		return NULL;

	if (list->iter == list->head) {
		list->iter = list->head->next;
	}

	AWNode *iter = list->iter;

	if (list->iter) {
		list->iter = list->iter->next;
	}

	return iter;
}
